using Microsoft.VisualBasic;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Diagnostics;
using System.Windows.Forms;

using System.Globalization;
//using System.Math;
namespace Theremino_HAL
{


	// ==================================================================================================
	//   CLASS SLAVE
	// ==================================================================================================
	class Slave
	{
		public enum SlaveTypes : int
		{
			Custom = 0,
			CapSensor = 1,
			Servo = 2,
			Generic = 3,
			InOut = 4,
			MasterPins = 5,
			//RmsLogger = 6,
			//Sniffer = 7,
            MasterPinsV2 = 8,
            MasterPinsV4 = 9,
			Unknown = 255
		}

		internal Int32 MasterId;
		internal Int32 mSlaveId;
		internal SlaveTypes SlaveType;
		internal List<Pin> Pins;

		internal Int32 LoadCurrent_mA;

		internal Int32 SlaveId {
			get { return mSlaveId; }
// test if mSlaveId ok
			set { mSlaveId = value; }
		}

		internal string SlaveTypeString()
		{
			switch (SlaveType) {
				case SlaveTypes.Custom:
					return "Custom";
				case SlaveTypes.CapSensor:
					return "CapSensor";
				case SlaveTypes.Servo:
					return "Servo";
				case SlaveTypes.Generic:
					return "Generic";
				case SlaveTypes.InOut:
					return "InOut";
				case SlaveTypes.MasterPins:
					return "MasterPins";
                case SlaveTypes.MasterPinsV2:
                    return "MasterPins";
                case SlaveTypes.MasterPinsV4:
                    return "MasterPins";
				case SlaveTypes.Unknown:
					return "Unknown";
				default:
					return "Unknown";
			}
		}


		internal Slave(SlaveTypes _slavetype, Int32 _masterid, Int32 _slaveid)
		{
			SlaveType = _slavetype;
			MasterId = _masterid;
			SlaveId = _slaveid;
			LoadCurrent_mA = 0;
		}

		internal Int32 GetHostToMasterBytesCount()
		{
			Int32 functionReturnValue = default(Int32);
			functionReturnValue = 0;
			foreach (Pin p in Pins) {
				if (p.Direction == Pin.Directions.HostToMaster) {
					functionReturnValue += p.UsbBytesCount;
				}
			}
			return functionReturnValue;
		}

		internal Int32 GetMasterToHostBytesCount()
		{
			Int32 functionReturnValue = default(Int32);
			functionReturnValue = 0;
			foreach (Pin p in Pins) {
				if (p.Direction == Pin.Directions.MasterToHost) {
					functionReturnValue += p.UsbBytesCount;
				}
			}
			return functionReturnValue;
		}

        internal bool CalibrationNeeded()
        {
            foreach (Pin p in Pins)
            {
                switch (p.GetPinType())
                {
                    case Pin.PinTypes.CAP_SENSOR:
                        if (p.Area_cmq > 0) return true;
                        break;
                    case Pin.PinTypes.CAP_8:
                    case Pin.PinTypes.CAP_16:
                        if (p.MinVariation > 0) return true;
                        break;
                }
            }
            return false;
        }

		internal virtual void Calibrate()
		{
			foreach (Pin p in Pins) {
				p.Calibrate();
			}
		}

		internal virtual void CalibrateZero()
		{
			foreach (Pin p in Pins) {
				p.CalibrateZero();
			}
		}

		internal virtual UInt32 Get_CapSensor_Or_Cap_Delta()
		{
			UInt32 maxdelta = 0;
            foreach (Pin p in Pins) 
            {
                switch (p.GetPinType())
                {
                    case Pin.PinTypes.CAP_8:
                    case Pin.PinTypes.CAP_16:
                    {
                        UInt32 n = Module_MathFunctions.AbsUint32Difference(p.Value_Uinteger_Max, p.Value_Uinteger_Min);
                        if (n > maxdelta) maxdelta = n;
                        p.ResetPosMinMax();
                    }
                    break;
                }
            }
            return maxdelta;
		}


		internal virtual void ConfigurePin(Int32 npin, Pin.PinTypes pinType)
		{
			switch (SlaveType) 
            {
                case SlaveTypes.Servo:
                case SlaveTypes.Generic:
                case SlaveTypes.InOut:
                {
                    switch (pinType)
                    {
                        case Pin.PinTypes.UNUSED:
                        case Pin.PinTypes.DIG_OUT:
                        case Pin.PinTypes.PWM_8:
                        case Pin.PinTypes.PWM_16:
                        case Pin.PinTypes.SERVO_8:
                        case Pin.PinTypes.SERVO_16:
                        case Pin.PinTypes.DIG_IN:
                        case Pin.PinTypes.DIG_IN_PU:
                        case Pin.PinTypes.COUNTER:
                        case Pin.PinTypes.COUNTER_PU:
                            Pins[npin].SetPinType(pinType);
                            break;
                        case Pin.PinTypes.ADC_8:
                        case Pin.PinTypes.ADC_16:
                        case Pin.PinTypes.CAP_8:
                        case Pin.PinTypes.CAP_16:
                        case Pin.PinTypes.RES_8:
                        case Pin.PinTypes.RES_16:
                            if (npin < 8) Pins[npin].SetPinType(pinType);
                            break;
                        case Pin.PinTypes.FAST_COUNTER:
                        case Pin.PinTypes.FAST_COUNTER_PU:
                            if (npin == 7) Pins[npin].SetPinType(pinType);
                            break;
                        case Pin.PinTypes.USOUND_SENSOR:
                        case Pin.PinTypes.PERIOD:
                        case Pin.PinTypes.PERIOD_PU:
                            if (npin == 8) Pins[npin].SetPinType(pinType);
                            break;
                    }
                    break;
                }
			}
		}

		internal virtual void FillTypeCombo(ComboBox cmb, ref Pin p)
		{
			switch (SlaveType) 
            {
				case SlaveTypes.Servo:
				case SlaveTypes.Generic:
				case SlaveTypes.InOut:
                {
					cmb.Items.Clear();
					//
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED));
					//
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_OUT));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_8));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_16));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_8));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_16));
					//
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN_PU));
					//
					if (p.PinId < 8) {
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_8));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_16));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_8));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_16));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_8));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_16));
					}

					//
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER));
					cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER_PU));

					if (p.PinId == 7) {
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER_PU));
					}

					if (p.PinId == 8) {
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD_PU));
						cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.USOUND_SENSOR));
					}
					break;
                }
				default:
					cmb.Items.Clear();
					cmb.Items.Add(Pin.PinTypeToString(p.GetPinType()));
					break;
			}
		}

        internal bool ValidLinkedToPrevious(Pin pinSelected)
        {
            if (pinSelected.GetPinType() != Pin.PinTypes.STEPPER) return false;
            if (pinSelected.PinId < 2) return false;
            if (Pins[pinSelected.PinId - 2].GetPinType() != Pin.PinTypes.STEPPER) return false;
            return false; // true;  // <<<<< TODO
        }
	}
}
namespace Theremino_HAL
{

	// ==================================================================================================
	// SUBCLASS - Slave_CapSensor
	// ==================================================================================================
	class Slave_CapSensor : Slave
	{

		internal Slave_CapSensor(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.CapSensor, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 12;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
		}

		internal override UInt32 Get_CapSensor_Or_Cap_Delta()
		{
			UInt32 functionReturnValue = default(UInt32);
			{
                UInt32 n = Module_MathFunctions.AbsUint32Difference(Pins[0].Value_Uinteger_Max, Pins[0].Value_Uinteger_Min);
				if (n < 0) n = 0; 
				if (n > 400000) n = 0; 
				functionReturnValue = n;
				Pins[0].ResetPosMinMax();
			}
			return functionReturnValue;
		}

		internal override void ConfigurePin(Int32 npin, Pin.PinTypes pinType)
		{
			switch (pinType) {
				case Pin.PinTypes.UNUSED:
				case Pin.PinTypes.CAP_SENSOR:
					Pins[npin].SetPinType(pinType);
					break;
			}
		}

		internal override void FillTypeCombo(ComboBox cmb, ref Pin p)
		{
			cmb.Items.Clear();
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_SENSOR));
		}

	}
}
namespace Theremino_HAL
{

	// ==================================================================================================
	//   SUBCLASS - Slave_Servo
	// ==================================================================================================
	class Slave_Servo : Slave
	{

		internal Slave_Servo(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.Servo, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 80;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9));
		}

	}
}

namespace Theremino_HAL
{
	// ==================================================================================================
	//   SUBCLASS - Slave_Generic
	// ==================================================================================================
	class Slave_Generic : Slave
	{

		internal Slave_Generic(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.Generic, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 80;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 10));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 11));
		}

	}
}

namespace Theremino_HAL
{
	// ==================================================================================================
	//   SUBCLASS - Slave_InOut
	// ==================================================================================================
	class Slave_InOut : Slave
	{

		internal Slave_InOut(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.InOut, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 80;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 10));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 11));
		}

	}
}

namespace Theremino_HAL
{
	// ==================================================================================================
	//   SUBCLASS - Slave_MasterPins
	// ==================================================================================================
	class Slave_MasterPins : Slave
	{

		internal Slave_MasterPins(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.MasterPins, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 10;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5));
		}

		internal override void ConfigurePin(Int32 npin, Pin.PinTypes pinType)
		{
			switch (pinType) 
            {
				case Pin.PinTypes.UNUSED:
				case Pin.PinTypes.DIG_OUT:
				case Pin.PinTypes.PWM_8:
				case Pin.PinTypes.PWM_16:
				case Pin.PinTypes.SERVO_8:
				case Pin.PinTypes.SERVO_16:
				case Pin.PinTypes.DIG_IN:
				case Pin.PinTypes.DIG_IN_PU:
				case Pin.PinTypes.ADC_8:
				case Pin.PinTypes.ADC_16:
				case Pin.PinTypes.CAP_8:
				case Pin.PinTypes.CAP_16:
				case Pin.PinTypes.RES_8:
				case Pin.PinTypes.RES_16:
				case Pin.PinTypes.COUNTER:
				case Pin.PinTypes.COUNTER_PU:
					Pins[npin].SetPinType(pinType);
					break;
				case Pin.PinTypes.FAST_COUNTER:
				case Pin.PinTypes.FAST_COUNTER_PU:
					if (!AlreadyUsed_FastCounter(Pins[npin])) Pins[npin].SetPinType(pinType); 
					break;
				case Pin.PinTypes.USOUND_SENSOR:
				case Pin.PinTypes.PERIOD:
				case Pin.PinTypes.PERIOD_PU:
					if (!AlreadyUsed_PeriodCounter(Pins[npin])) Pins[npin].SetPinType(pinType); 
					break;
			}
		}

		internal override void FillTypeCombo(ComboBox cmb, ref Pin p)
		{
			cmb.Items.Clear();
			//
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED));
			//
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_OUT));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_8));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_16));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_8));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_16));
			//
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN_PU));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_8));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_16));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_8));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_16));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_8));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_16));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER));
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER_PU));

			if (!AlreadyUsed_FastCounter(p)) 
            {
				cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER));
				cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER_PU));
			}

			if (!AlreadyUsed_PeriodCounter(p)) 
            {
				cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD));
				cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD_PU));
				cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.USOUND_SENSOR));
			}
		}

		private bool AlreadyUsed_FastCounter(Pin pinSelected)
		{
			foreach (Pin p in Pins) {
				if (object.ReferenceEquals(p, pinSelected)) continue; 
				if (p.GetPinType() == Pin.PinTypes.FAST_COUNTER || p.GetPinType() == Pin.PinTypes.FAST_COUNTER_PU) {
					return true;
				}
			}
			return false;
		}

		private bool AlreadyUsed_PeriodCounter(Pin pinSelected)
		{
			foreach (Pin p in Pins) {
				if (object.ReferenceEquals(p, pinSelected)) continue; 
				if (p.GetPinType() == Pin.PinTypes.PERIOD || p.GetPinType() == Pin.PinTypes.PERIOD_PU || p.GetPinType() == Pin.PinTypes.USOUND_SENSOR) {
					return true;
				}
			}
			return false;
		}
	}
}

namespace Theremino_HAL
{
	// ==================================================================================================
	//   SUBCLASS - Slave_MasterPinsV2
	// ==================================================================================================
	class Slave_MasterPinsV2 : Slave
	{

		internal Slave_MasterPinsV2(Int32 _masterid, Int32 _slaveid) : base(SlaveTypes.MasterPins, _masterid, _slaveid)
		{
			InitDefaultPins();
			LoadCurrent_mA = 10;
		}

		private void InitDefaultPins()
		{
			Pins = new List<Pin>();
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4));
			Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9));
		}

		internal override void ConfigurePin(Int32 npin, Pin.PinTypes pinType)
		{
			switch (pinType) {

				case Pin.PinTypes.UNUSED:
                    Pins[npin].SetPinType(pinType);
                    break;

				case Pin.PinTypes.DIG_OUT:
				case Pin.PinTypes.PWM_8:
				case Pin.PinTypes.PWM_16:
				case Pin.PinTypes.SERVO_8:
				case Pin.PinTypes.SERVO_16:
				case Pin.PinTypes.DIG_IN:
				case Pin.PinTypes.DIG_IN_PU:
				case Pin.PinTypes.COUNTER:
				case Pin.PinTypes.COUNTER_PU:
                case Pin.PinTypes.SLOW_PERIOD:
                case Pin.PinTypes.SLOW_PERIOD_PU:
                    if (ValidGenericPin(Pins[npin])) Pins[npin].SetPinType(pinType);
					break;

                case Pin.PinTypes.ADC_8:
                case Pin.PinTypes.ADC_16:
                case Pin.PinTypes.CAP_8:
                case Pin.PinTypes.CAP_16:
                case Pin.PinTypes.RES_8:
                case Pin.PinTypes.RES_16:
                    if (ValidAdcPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.STEPPER:
                    if (ValidStepperPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.STEPPER_DIR:
                    if (ValidStepperDirPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.PWM_FAST:
                    if (ValidPwmFastPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

				case Pin.PinTypes.FAST_COUNTER:
				case Pin.PinTypes.FAST_COUNTER_PU:
					if (ValidFastCounterPin(Pins[npin])) Pins[npin].SetPinType(pinType); 
					break;

				case Pin.PinTypes.USOUND_SENSOR:
				case Pin.PinTypes.PERIOD:
				case Pin.PinTypes.PERIOD_PU:
					if (ValidPeriodCounterPin(Pins[npin])) Pins[npin].SetPinType(pinType); 
					break;
			}
		}

		internal override void FillTypeCombo(ComboBox cmb, ref Pin p)
		{
			cmb.Items.Clear();
			
			cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED));
			
            if (ValidGenericPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_OUT));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_8));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_16));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_8));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_16));
            }

            if (ValidStepperPin(p))
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.STEPPER));
            if (ValidStepperDirPin(p))
            {
                cmb.Items.Clear();
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.STEPPER_DIR));
            }
            if (ValidPwmFastPin(p))
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_FAST));

            if (ValidGenericPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN_PU));
            }

            if (ValidAdcPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_8));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_16));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_8));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_16));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_8));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_16));
            }

            if (ValidGenericPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER_PU));
            }

            if (ValidFastCounterPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER_PU));
            }

            if (ValidPeriodCounterPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD_PU));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.USOUND_SENSOR));
            }

            if (ValidGenericPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SLOW_PERIOD));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SLOW_PERIOD_PU));
            }
		}

        private bool ValidGenericPin(Pin pinSelected)
        {
            if (pinSelected.PinId > 0)
            {
                if (Pins[pinSelected.PinId - 1].GetPinType() == Pin.PinTypes.STEPPER)
                {
                    return false;
                }
            }
            return true;
        }

        private bool ValidAdcPin(Pin pinSelected)
        {
            if (!ValidGenericPin(pinSelected)) return false;
            if (pinSelected.PinId > 5) return false;
            return true;
        }

        private bool ValidStepperPin(Pin pinSelected)
        {
            if (CountUsedPwmPins(pinSelected) > 4) return false;
            switch (pinSelected.PinId)
            {
                case 0:
                case 2:
                case 4:
                case 6:
                case 8:
                    return true;
            }
            return false;
        }

        private bool ValidStepperDirPin(Pin pinSelected)
        {
            if (pinSelected.PinId < 1) return false;
            if (Pins[pinSelected.PinId - 1].GetPinType() != Pin.PinTypes.STEPPER) return false;
            return true;
        }

        private bool ValidPwmFastPin(Pin pinSelected)
        {
            if (!ValidGenericPin(pinSelected)) return false;
            if (CountUsedPwmPins(pinSelected) > 4) return false;
            return true;
        }

		private bool ValidFastCounterPin(Pin pinSelected)
		{
            if (!ValidGenericPin(pinSelected)) return false;
			foreach (Pin p in Pins) 
            {
				if (object.ReferenceEquals(p, pinSelected)) continue; 
				if (p.GetPinType() == Pin.PinTypes.FAST_COUNTER || p.GetPinType() == Pin.PinTypes.FAST_COUNTER_PU) {
					return false;
				}
			}
			return true;
		}

		private bool ValidPeriodCounterPin(Pin pinSelected)
		{
            if (!ValidGenericPin(pinSelected)) return false;
			foreach (Pin p in Pins) 
            {
				if (object.ReferenceEquals(p, pinSelected)) continue; 
				if (p.GetPinType() == Pin.PinTypes.PERIOD || p.GetPinType() == Pin.PinTypes.PERIOD_PU || 
                                                             p.GetPinType() == Pin.PinTypes.USOUND_SENSOR) 
                {
					return false;
				}
			}
			return true;
		}

        private Int32 CountUsedPwmPins(Pin pinSelected)
        {
            Int32 n = 0;
            foreach (Pin p in Pins)
            {
                if (object.ReferenceEquals(p, pinSelected)) continue;
                if (p.GetPinType() == Pin.PinTypes.STEPPER || p.GetPinType() == Pin.PinTypes.PWM_FAST)
                {
                    n++;
                }
            }
            return n;
        }

	}
}


namespace Theremino_HAL
{
    // ==================================================================================================
    //   SUBCLASS - Slave_MasterPinsV4
    // ==================================================================================================
    class Slave_MasterPinsV4 : Slave
    {

        internal Slave_MasterPinsV4(Int32 _masterid, Int32 _slaveid)
            : base(SlaveTypes.MasterPins, _masterid, _slaveid)
        {
            InitDefaultPins();
            LoadCurrent_mA = 10;
        }

        private void InitDefaultPins()
        {
            Pins = new List<Pin>();
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 0));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 1));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 2));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 3));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 4));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 5));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 6));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 7));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 8));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 9));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 10));
            Pins.Add(new Pin(Pin.PinTypes.UNUSED, MasterId, mSlaveId, 11));
        }

        internal override void ConfigurePin(Int32 npin, Pin.PinTypes pinType)
        {
            switch (pinType)
            {

                case Pin.PinTypes.UNUSED:
                    Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.DIG_OUT:
                case Pin.PinTypes.PWM_8:
                case Pin.PinTypes.PWM_16:
                case Pin.PinTypes.SERVO_8:
                case Pin.PinTypes.SERVO_16:
                case Pin.PinTypes.DIG_IN:
                case Pin.PinTypes.DIG_IN_PU:
                case Pin.PinTypes.COUNTER:
                case Pin.PinTypes.COUNTER_PU:
                case Pin.PinTypes.SLOW_PERIOD:
                case Pin.PinTypes.SLOW_PERIOD_PU:
                    if (ValidGenericPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.ADC_8:
                case Pin.PinTypes.ADC_16:
                case Pin.PinTypes.CAP_8:
                case Pin.PinTypes.CAP_16:
                case Pin.PinTypes.RES_8:
                case Pin.PinTypes.RES_16:
                    if (ValidAdcPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.STEPPER:
                    if (ValidStepperPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.STEPPER_DIR:
                    if (ValidStepperDirPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.PWM_FAST:
                    if (ValidPwmFastPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.FAST_COUNTER:
                case Pin.PinTypes.FAST_COUNTER_PU:
                    if (ValidFastCounterPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.USOUND_SENSOR:
                case Pin.PinTypes.PERIOD:
                case Pin.PinTypes.PERIOD_PU:
                    if (ValidPeriodCounterPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.ENCODER_A:
                case Pin.PinTypes.ENCODER_A_PU:
                    if (ValidEncoderAPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;

                case Pin.PinTypes.ENCODER_B:
                case Pin.PinTypes.ENCODER_B_PU:
                    if (ValidEncoderBPin(Pins[npin])) Pins[npin].SetPinType(pinType);
                    break;
            }
        }

        internal override void FillTypeCombo(ComboBox cmb, ref Pin p)
        {
            cmb.Items.Clear();

            cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.UNUSED));

            if (ValidGenericPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_OUT));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_8));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_16));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_8));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SERVO_16));
            }

            if (ValidStepperPin(p))
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.STEPPER));
            if (ValidStepperDirPin(p))
            {
                cmb.Items.Clear();
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.STEPPER_DIR));
            }
            if (ValidPwmFastPin(p))
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PWM_FAST));

            if (ValidGenericPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.DIG_IN_PU));
            }

            if (ValidAdcPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_8));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ADC_16));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_8));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.CAP_16));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_8));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.RES_16));
            }

            if (ValidGenericPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.COUNTER_PU));
            }

            if (ValidFastCounterPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.FAST_COUNTER_PU));
            }

            if (ValidPeriodCounterPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.PERIOD_PU));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.USOUND_SENSOR));
            }

            if (ValidEncoderAPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ENCODER_A));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ENCODER_A_PU));
            }
            if (ValidEncoderBPin(p))
            {
                cmb.Items.Clear();
                if (Pins[p.PinId - 1].GetPinType() == Pin.PinTypes.ENCODER_A) 
                {
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ENCODER_B));
                }
                else
                {
                    cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.ENCODER_B_PU));
                }
            }

            if (ValidGenericPin(p))
            {
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SLOW_PERIOD));
                cmb.Items.Add(Pin.PinTypeToString(Pin.PinTypes.SLOW_PERIOD_PU));
            }
        }

        private bool ValidGenericPin(Pin pinSelected)
        {
            if (pinSelected.PinId > 0)
            {
                if (Pins[pinSelected.PinId - 1].GetPinType() == Pin.PinTypes.STEPPER)
                {
                    return false;
                }
                if (Pins[pinSelected.PinId - 1].GetPinType() == Pin.PinTypes.ENCODER_A)
                {
                    return false;
                }
                if (Pins[pinSelected.PinId - 1].GetPinType() == Pin.PinTypes.ENCODER_A_PU)
                {
                    return false;
                }
            }
            return true;
        }

        private bool ValidAdcPin(Pin pinSelected)
        {
            if (!ValidGenericPin(pinSelected)) return false;
            if (pinSelected.PinId > 5) return false;
            return true;
        }

        private bool ValidStepperPin(Pin pinSelected)
        {
            if (CountUsedPwmPins(pinSelected) > 4) return false;
            switch (pinSelected.PinId)
            {
                case 0:
                case 2:
                case 4:
                case 6:
                case 8:
                    return true;
            }
            return false;
        }

        private bool ValidStepperDirPin(Pin pinSelected)
        {
            if (pinSelected.PinId < 1) return false;
            if (Pins[pinSelected.PinId - 1].GetPinType() != Pin.PinTypes.STEPPER) return false;
            return true;
        }

        private bool ValidPwmFastPin(Pin pinSelected)
        {
            if (pinSelected.PinId > 9) return false;
            if (!ValidGenericPin(pinSelected)) return false;
            if (CountUsedPwmPins(pinSelected) > 4) return false;
            return true;
        }

        private bool ValidFastCounterPin(Pin pinSelected)
        {
            if (pinSelected.PinId > 9) return false;
            if (!ValidGenericPin(pinSelected)) return false;
            foreach (Pin p in Pins)
            {
                if (object.ReferenceEquals(p, pinSelected)) continue;
                if (p.GetPinType() == Pin.PinTypes.FAST_COUNTER || p.GetPinType() == Pin.PinTypes.FAST_COUNTER_PU)
                {
                    return false;
                }
            }
            return true;
        }

        private bool ValidPeriodCounterPin(Pin pinSelected)
        {
            if (pinSelected.PinId > 9) return false;
            if (!ValidGenericPin(pinSelected)) return false;
            foreach (Pin p in Pins)
            {
                if (object.ReferenceEquals(p, pinSelected)) continue;
                if (p.GetPinType() == Pin.PinTypes.PERIOD || p.GetPinType() == Pin.PinTypes.PERIOD_PU ||
                                                             p.GetPinType() == Pin.PinTypes.USOUND_SENSOR)
                {
                    return false;
                }
            }
            return true;
        }

        private bool ValidEncoderAPin(Pin pinSelected)
        {
            if ((pinSelected.PinId > 10)) return false;
            Pin.PinTypes pt = Pins[pinSelected.PinId + 1].GetPinType();

            if (pt == Pin.PinTypes.STEPPER) return false;
            if (pt == Pin.PinTypes.ENCODER_A) return false;
            if (pt == Pin.PinTypes.ENCODER_A_PU) return false;
            if (pinSelected.PinId > 0)
            {
                pt = Pins[pinSelected.PinId - 1].GetPinType();
                if (pt == Pin.PinTypes.STEPPER) return false;
                if (pt == Pin.PinTypes.ENCODER_A) return false;
                if (pt == Pin.PinTypes.ENCODER_A_PU) return false;
            }
            return true;
        }

        private bool ValidEncoderBPin(Pin pinSelected)
        {
            if (pinSelected.PinId < 1) return false;

            Pin.PinTypes pt = Pins[pinSelected.PinId - 1].GetPinType();

                if (pt != Pin.PinTypes.ENCODER_A
                    && 
                    pt != Pin.PinTypes.ENCODER_A_PU)
                {
                    return false;
                }
            return true;
        }

        private Int32 CountUsedPwmPins(Pin pinSelected)
        {
            Int32 n = 0;
            foreach (Pin p in Pins)
            {
                if (object.ReferenceEquals(p, pinSelected)) continue;
                if (p.GetPinType() == Pin.PinTypes.STEPPER || p.GetPinType() == Pin.PinTypes.PWM_FAST)
                {
                    n++;
                }
            }
            return n;
        }

    }
}


